#!/usr/bin/python3

import sys
sys.path.insert(0, "../../../Python")

import rumd
# it is essential that there is no doubt about which version is actually run!
print("File used to import the rumd module", rumd.__file__)

from rumd.Simulation import Simulation
from rumd.RUMD_Test import RUMD_Test

class Test_IHS(RUMD_Test):
    """
    Inherent state
    """
    def __init__(self):
        RUMD_Test.__init__(self)
        self.numParticlesList = [512, 1000, 8192]
        self.energy_tol["max_force"] = 0.01

    def CompareOutputDirect(self):
        """ Return non-zero if differences occur in output files"""
        num_diffs = RUMD_Test.CompareOutputDirect(self)
        num_diffs += self.CompareTrajectoryFiles("ihs")
        num_diffs += self.CompareEnergiesFiles("ihsEnergies", "ihs")
        return num_diffs

    def RunSimulation(self, mode, numParticles, pb, tp):

        # set the number of time-steps depending on the mode (assigned
        # as a class attribute so the calling function can access it later)
        self.numTimeSteps = {"short":         100,
                            "performance":  5000,
                            "stress":    1000000}[mode]

        # sample and simulation
        startConfg = "startN%d.xyz.gz" % numParticles


        sim = Simulation(startConfg, pb, tp, verbose=False)

        if self.generate_reference_data:
            sim.SetOutputDirectory(self.ref_directory)
        else:
            sim.SetOutputDirectory(self.out_directory)

        # Create LJ pair-potential
        potential = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        # Set up parameters for Lennard-Jones interactions ( Kob & Andersen )
        # i  j   Sig   Eps  Rcut (in units of Sigma_ij)
        potential.SetParams(0, 0, 1.00, 1.00, 2.5)
        potential.SetParams(0, 1, 0.80, 1.50, 2.5)
        potential.SetParams(1, 1, 0.88, 0.50, 2.5)
        sim.AddPotential(potential)

        # integrator
        itg = rumd.IntegratorNVT(timeStep=0.004, targetTemperature=0.8, thermostatRelaxationTime=0.2)
        sim.SetIntegrator(itg)

        ihs_man = rumd.IHS_OutputManager(sample=sim.sample, timeStep=0.015)
        ihs_man.SetMaximumNumberIterations(5000)
        sim.AddOutputManager("ihs", ihs_man)

        # tune output depending on the mode
        if mode == "short":
            sim.SetOutputScheduling("energies", "linear", interval=1)
            sim.SetOutputScheduling("trajectory", "linear", interval=10)
            sim.SetOutputScheduling("ihs", "linear", interval=10)
            sim.SetBlockSize(50)
        elif mode == "performance":
            sim.SetOutputScheduling("energies", "none")
            sim.SetOutputScheduling("trajectory", "none")
            ihs_man.SetWriteIHS_Config(False)
            sim.SetBlockSize(1000)
            sim.SetOutputScheduling("ihs", "linear", interval=100)
        else:
            assert mode == "stress"
            sim.SetOutputScheduling("energies", "linear", interval=500)
            # logarithmic saving for trajectory and quenches

        # want output also for performance because otherwise IHS does not get
        # computed, and we wouldn't be measuring its performance
        sim.Run(self.numTimeSteps)
        sim.sample.TerminateOutputManagers()

if __name__ == "__main__":

    rumd_test = Test_IHS()

    # standard short test to be called from makefile
    # For generating reference data and running stress and performance modes
    # there are corresponding scripts to be submitted to pbs

    sys.exit(rumd_test.RunShort("gen_ref" in sys.argv))

    # for individual runs, could do for example
    #sys.exit(rumd_test.Test("short", 1000, 16, 8))
