#!/usr/bin/python3

import os
if "PBS_O_WORKDIR" in os.environ:
    workingDir = os.environ["PBS_O_WORKDIR"]
    os.chdir(workingDir)

import sys
sys.path.insert(0, "../../../Python")

import rumd
# it is essential that there is no doubt about which version is actually run!
print("File used to import the rumd module", rumd.__file__)

from rumd.Simulation import Simulation
from rumd.RUMD_Test import RUMD_Test

class Test_MMC_KobAndersen(RUMD_Test):
    """
    Binary LJ system with MC-NVT dynamics.
    """
    def __init__(self):
        RUMD_Test.__init__(self)

    def RunSimulation(self, mode, numParticles, pb, tp):

        # set the number of time-steps depending on the mode
        self.numTimeSteps = {"short":          100,
                             "performance":100000,
                             "stress":   100000000}[mode]

        # sample and simulation
        startConfg = "startN%d.xyz.gz" % numParticles


        sim = Simulation(startConfg, pb, tp, verbose=False)

        if self.generate_reference_data:
            sim.SetOutputDirectory(self.ref_directory)
        else:
            sim.SetOutputDirectory(self.out_directory)

        # integrator
        itg = rumd.IntegratorMMC(dispLength=0.005, targetTemperature=0.8)

        # Create LJ pair-potential
        potential = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        # Set up parameters for Lennard-Jones interactions ( Kob & Andersen )
        # i  j   Sig   Eps  Rcut (in units of Sigma_ij)
        potential.SetParams(0, 0, 1.00, 1.00, 2.5)
        potential.SetParams(0, 1, 0.80, 1.50, 2.5)
        potential.SetParams(1, 1, 0.88, 0.50, 2.5)

        sim.SetIntegrator(itg)
        sim.AddPotential(potential)

        itg.SetHybridTausSeeds(0, 1, 2, 130) # Bad seeding but need to testing.

        # tune output depending on the mode
        if mode == "short":
            sim.SetOutputScheduling("energies", "linear", interval=1)
            sim.SetOutputScheduling("trajectory", "linear", interval=10)
            sim.SetBlockSize(50)
        elif mode == "stress":
            sim.SetOutputScheduling("energies", "linear", interval=5000)
        else:
            assert mode == "performance"
            print("Note there is no output directory for performance" \
                  " test--the only output is the timing file")

        sim.Run(self.numTimeSteps,
                suppressAllOutput=(mode == "performance"))



if __name__ == "__main__":
    rumd_test = Test_MMC_KobAndersen()

    # standard short test to be called from makefile
    # For generating reference data and running stress and performance modes
    # there are corresponding scripts to be submitted to pbs

    sys.exit(rumd_test.RunShort("gen_ref" in sys.argv))

    # for individual runs, could do for example
    #sys.exit(rumd_test.Test("short", 1000, 16, 8))
