#!/usr/bin/python3

import sys
sys.path.insert(0, "../../../Python")

import rumd
# it is essential that there is no doubt about which version is actually run!
print("File used to import the rumd module", rumd.__file__)

from rumd.Simulation import Simulation
from rumd.RUMD_Test import RUMD_Test

class Test_NPT_LJ(RUMD_Test):
    """
    Single-component LJ system with NPT dynamics.
    """
    def __init__(self):
        RUMD_Test.__init__(self)
        # only one size available, and this uses a special
        # configuration designed to reproduce the NPT-bug present up to
        # version 1610
        self.numParticlesList = [512]

        self.brelax = 10
        self.trelax = 0.4

    def RunSimulation(self, mode, numParticles, pb, tp):
        """ Do one simulation run for a given system size, mode, pb, tp"""
        # set the number of time-steps depending on the mode (assigned
        # as a class attribute so the calling function can access it later)
        self.numTimeSteps = {"short":        100,
                             "performance": 100000,
                             "stress":   100000000}[mode]

        if numParticles != 512:
            raise ValueError("test_NPT_LJ only available forN=512")
        sim = Simulation("startN512.xyz.gz", pb, tp, verbose=False)

        if self.generate_reference_data:
            sim.SetOutputDirectory(self.ref_directory)
        else:
            sim.SetOutputDirectory(self.out_directory)


        # Create LJ pair-potential
        potential = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        # Set up parameters for Lennard-Jones interactions
        potential.SetParams(0, 0, 1.00, 1.00, 2.5)
        if not self.generate_reference_data:
            potential.SetNB_Method("sort")

        # integrator
        itg = rumd.IntegratorNPTAtomic(timeStep=0.0025, targetTemperature=0.6,
                                       thermostatRelaxationTime=self.trelax,
                                       targetPressure=-0.212,
                                       barostatRelaxationTime=self.brelax)

        sim.SetIntegrator(itg)
        sim.AddPotential(potential)


        sim.SetOutputMetaData("energies", volume=True)

        # tune output depending on the mode
        if mode == "short":
            sim.SetOutputScheduling("energies", "linear", interval=1)
            sim.SetOutputScheduling("trajectory", "linear", interval=10)
            sim.SetBlockSize(50)
        elif mode == "stress":
            sim.SetOutputScheduling("energies", "linear", interval=5000)
        else:
            assert mode == "performance"
            print("Note there is no output directory for performance" \
                  " test--the only output is the timing file")

        sim.Run(self.numTimeSteps,
                suppressAllOutput=(mode == "performance"))


if __name__ == "__main__":

    rumd_test = Test_NPT_LJ()

    # standard short test to be called from makefile
    # For generating reference data and running stress and performance modes
    # there are corresponding scripts to be submitted to pbs

    sys.exit(rumd_test.RunShort("gen_ref" in sys.argv))
