#!/usr/bin/python3

"""
This script calculates the free energy of the WCA system along the T=0.80
isotherm by thermodynamic integration. It does not fit into the scheme used
by the other tests.
"""

import sys
import os
import numpy

from scipy.optimize import curve_fit

sys.path.insert(0, "../../../Python")

import rumd
# it is essential that there is no doubt about which version is actually run!
print("File used to import the rumd module", rumd.__file__)
from rumd.Simulation import Simulation
import rumd.Tools


if "PBS_O_WORKDIR" in os.environ:
    workingDir = os.environ["PBS_O_WORKDIR"]
    os.chdir(workingDir)
    sys.path.append(".")



class WCA_FreeEnergy(object):
    """ Class for implementing free energy calculations """
    def __init__(self, rhoMin=0.05, rhoMax=0.90, rhoStep=0.05,
                 startConfigFile="start.xyz.gz", LJcutoff=4.0):
        self.timeStep = 0.002
        self.rhoMin = rhoMin
        self.rhoMax = rhoMax
        self.rhoStep = rhoStep
        self.potLJCutoff = LJcutoff


        self.n_equil_steps = 25000
        self.n_run_steps = 50000
        self.momentum_reset_interval = 5000

        self.n_equil_stepsStart = 100000
        self.energyInterval = 50

        self.potWCA = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        self.potWCA.SetParams(0, 0, 1., 1., pow(2., 1./6))
        self.potLJ = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        self.potLJ.SetParams(0, 0, 1., 1., self.potLJCutoff)
        self.vLJ_cut = 4*(1/self.potLJCutoff**12 - 1/self.potLJCutoff**6)
        self.sim = Simulation(startConfigFile, 32, 8)
        self.sim.AddPotential(self.potWCA)
        self.T = None

        self.alt_pot_calc = None

    def Initialize(self):
        """ Create simulation object, integrator, potential, etc"""


        itg = rumd.IntegratorNVT(targetTemperature=self.T, timeStep=self.timeStep, thermostatRelaxationTime=0.2)
        self.sim.SetIntegrator(itg)

        self.SetDensity(self.rhoMin)


        self.sim.SetOutputScheduling("trajectory", "none")
        self.sim.SetOutputScheduling("energies", "linear", interval=self.energyInterval)
        self.sim.momentum_reset_interval = self.momentum_reset_interval

        self.sim.SetVerbose(False)

        self.sim.Run(self.n_equil_stepsStart, suppressAllOutput=True)

        self.alt_pot_calc = rumd.AlternatePotentialCalculator(self.sim.sample, self.potLJ)
        self.sim.AddExternalCalculator(self.alt_pot_calc)




    def SetDensity(self, rh):
        """ scale simulation to get density rh"""
        nParticles = self.sim.GetNumberOfParticles()
        vol = self.sim.GetVolume()
        currentDensity = nParticles/vol
        scaleFactor = pow(rh/currentDensity, -1./3)

        self.sim.ScaleSystem(scaleFactor)



    def MainLoop(self, T):
        """ For given temperature T, does looop over densities to get
        data for thermodynamic integration"""
        self.T = T
        nRuns = int((self.rhoMax - self.rhoMin)/self.rhoStep + 1.e-8) + 1

        self.Initialize()

        outfile = open("outputWCA%.3f.dat" % T, "w")
        outfile.write("# rho, meanW, W0, meanPE\n")

        for rdx in range(nRuns):

            rho = self.rhoMin + rdx*self.rhoStep
            print("rdx, rho", rdx, rho)
            self.SetDensity(rho)

            self.sim.Run(self.n_equil_steps, suppressAllOutput=True)

            self.sim.Run(self.n_run_steps)


            rs = rumd.Tools.rumd_stats()
            rs.ComputeStats()
            meanVals = rs.GetMeanVals()
            meanPE = meanVals["pe"]
            meanW = meanVals["W"]
            if "potLJ_12_6" in meanVals:
                mean_pe_LJ = meanVals["potLJ_12_6"]

                correction1 = 0.5*(4.*numpy.pi/3*self.potLJCutoff**3*rho - 1)*self.vLJ_cut
                correction2 = -8.*numpy.pi*rho/3./self.potLJCutoff**3
                mean_pe_LJ += correction1 + correction2

                W0 = mean_pe_LJ - meanPE
            else:
                W0 = 0.0

            outfile.write("%.3f %.5f %.5f %.5f\n" % (rho, meanW, W0, meanPE))
        outfile.close()


def PolyModel2(rh, *params):
    """Polynomial function for fitting for use with curve_fit from scipy

    The degree is determined by size of params
    but coefficients 0 and 1 are not included

    arguments:
    rh -- the independent variable of the function, here a density (rho)
    params -- a variable number of parameters, in this case poynomial
              coefficients starting from 2

    """
    r = 0.0
    e = 2
    for p in params:
        r += p*numpy.power(rh, e)
        e += 1
    return r



if __name__ == "__main__":


    print("##############################################################" \
          "######\n\nNote: WCA_FreeEnergy takes a few minutes to run\n\n###" \
          "#################################################################")

    gen_ref = (len(sys.argv) > 1 and sys.argv[1] == "gen_ref")

    T = 0.80
    WCA = WCA_FreeEnergy()
    WCA.rhoMin = 0.05
    WCA.rhoMax = 0.50
    WCA.rhoStep = 0.05

    WCA.MainLoop(T)


    nParams = 6
    # second and third virial coefficients of WCA potential at T=0.80
    B2, B3 = 2.25566, 3.15556

    infile = open("outputWCA%.3f.dat" % T)


    W0List = []
    toFitList = []
    nextLine = infile.readline()
    while nextLine:
        while nextLine.startswith("#"):
            nextLine = infile.readline()
        rho, meanW, W0, meanPE = [float(item) for item in nextLine.split()]
        minusKnownVir = meanW/T/rho - B2 - B3*rho
        toFitList.append([rho, minusKnownVir])
        W0List.append(W0)

        nextLine = infile.readline()

    # fit this to polynomial starting with rho**2
    startParams = numpy.zeros(nParams)

    #for pdx in xrange(nParams):
    #    startParams.append(0.)

    toFitData = numpy.array(toFitList)
    fitParams, fitCov = curve_fit(PolyModel2, toFitData[:, 0], toFitData[:, 1], startParams)


    if gen_ref:
        f_outfile = open("free_energy_over_T_ref.dat", "w")
    else:
        f_outfile = open("free_energy_over_T.dat", "w")

    f_outfile.write("# rho total_F_over_T\n")

    if not gen_ref:
        sum_sq_diffs = 0
        infile = open("free_energy_over_T_ref.dat")
        infile.readline() # get the comment line out of the way

    #for rdx in xrange(len(toFitList)):
    for rho, W0 in zip(toFitData[:, 0], W0List):
        #rho, valToFit = toFitList[rdx]
        fittedVal = PolyModel2(rho, fitParams)
        integral = B2*rho + B3*rho**2/2
        for pdx in range(nParams):
            integral += fitParams[pdx] * rho**(3+pdx)/(3.+pdx)

        totalFoverT = integral + W0/T
        f_outfile.write("%f %f\n" % (rho, totalFoverT))

        if not gen_ref:
            ref_rho, ref_FoverT = [float(item) for item in
                                   infile.readline().split()]
            sum_sq_diffs += (totalFoverT - ref_FoverT)**2
    f_outfile.close()

    if not gen_ref:
        print("sum_sq_diffs", sum_sq_diffs)
        if sum_sq_diffs < 1.e-3:
            sys.exit(0)
        else:
            sys.exit(1)
