"""
Adapters for RUMD simulation package
"""

import os



class rumdSimulation(object):

    def __init__(self, sim, trajectory):
        self.trajectory = trajectory
        self._sim = sim
        self._sim.SetMomentumResetInterval(10)
        self._verbosity = 0
        self._suppress_all_output = True

    def _set_verbosity(self, value):
        self._verbosity = value
        if value == 0:
            self._sim.SetVerbose(False)
            self._sim.SetOutputScheduling("energies", "none")
            self._sim.SetOutputScheduling("trajectory", "none")
            self._suppress_all_output = True
        else:
            self._sim.SetVerbose(True)
            self._suppress_all_output = False
            # TODO: reset output parameters, or get rid of it

    def _get_verbosity(self):
        return self._verbosity

    verbosity = property(_get_verbosity, _set_verbosity, 'Verbosity')

    def _get_system(self):
        return System(self._sim)

    def _set_system(self, value):
        self._sim.sample = value.sample

    system = property(_get_system, _set_system, 'System')

    def run(self, n):
        self._sim.Run(n, suppressAllOutput=self._suppress_all_output)


class System(object):

    def __init__(self, sim):
        self._sim = sim
        self.sample = sim.sample

    def potential_energy(self):
        self.sample.CalcF()
        return self.sample.GetPotentialEnergy()

    def temperature(self):
        npart = self.sample.GetNumberOfParticles()
        vel = self.sample.GetVelocities()
        T = sum(sum(vel**2.0))/(3*npart - 3)
        return T

    def mean_square_displacement(self, reference):
        """ Compute the mean square displacement between actual sample and the reference sample """
        ndim = 3 # hard coded
        N = self.sample.GetNumberOfParticles()
        L = [self.sample.GetSimulationBox().GetLength(i) for i in range(ndim)]

        # Unfold positions using periodic image information
        ref = reference.sample.GetPositions() + reference.sample.GetImages() * L
        unf = self.sample.GetPositions() + self.sample.GetImages() * L

        return sum(sum((unf-ref)**2)) / N

    def _get_thermostat(self):
        return self.sample.GetIntegrator()

    def _set_thermostat(self, value):
        self._sim.SetIntegrator(value)

    thermostat = property(_get_thermostat,
                          _set_thermostat,
                          'Thermostat (actually an instance ' \
                          'of RUMD Integrator')


class Trajectory(object):

    def __init__(self, filename, mode='r'):
        self.filename = filename
        self.mode = 'w'

        # Remove the .gz extension from path. It will be added by rumd anyway
        base, ext = os.path.splitext(self.filename)
        if ext == '.gz':
            self.filename = base

    def write_initial_state(self, system):
        pass

    def write_sample(self, system, step, sample):
        system.sample.WriteConf(self.filename, self.mode)

        # Next time we append
        if self.mode == 'w':
            self.mode = 'a'

    def close(self):
        # Assuming something has been written, unzip the trajectory file
        if os.path.exists(self.filename + '.gz'):
            os.system("gunzip -f %s.gz" % self.filename)
